# Copyright 2014 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the
#    License. You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing,
#    software distributed under the License is distributed on an "AS
#    IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#    express or implied. See the License for the specific language
#    governing permissions and limitations under the License.

import unittest

import mock

from tugboat import pulls


class PullRequestTest(unittest.TestCase):
    @mock.patch.object(pulls.PullRequest, '__init__', return_value=None)
    def test_from_repos_no_callback(self, mock_init):
        prs1 = ['pr1_1', 'pr1_2', 'pr1_3']
        prs2 = ['pr2_1', 'pr2_2']
        repo1 = mock.Mock(**{'get_pulls.return_value': prs1})
        repo2 = mock.Mock(**{'get_pulls.return_value': prs2})

        result = pulls.PullRequest._from_repos(
            (r for r in (repo1, repo2)), None)

        self.assertEqual(len(result), 5)
        for pr in result:
            self.assertTrue(isinstance(pr, pulls.PullRequest))
        repo1.get_pulls.assert_called_once_with()
        repo2.get_pulls.assert_called_once_with()
        mock_init.assert_has_calls([
            mock.call(repo1, 'pr1_1'),
            mock.call(repo1, 'pr1_2'),
            mock.call(repo1, 'pr1_3'),
            mock.call(repo2, 'pr2_1'),
            mock.call(repo2, 'pr2_2'),
        ])
        self.assertEqual(mock_init.call_count, 5)

    @mock.patch.object(pulls.PullRequest, '__init__', return_value=None)
    def test_from_repos_with_callback(self, mock_init):
        prs1 = ['pr1_1', 'pr1_2', 'pr1_3']
        prs2 = ['pr2_1', 'pr2_2']
        repo1 = mock.Mock(**{'get_pulls.return_value': prs1})
        repo2 = mock.Mock(**{'get_pulls.return_value': prs2})
        cb = mock.Mock()

        result = pulls.PullRequest._from_repos(
            (r for r in (repo1, repo2)), cb)

        self.assertEqual(len(result), 5)
        for pr in result:
            self.assertTrue(isinstance(pr, pulls.PullRequest))
        repo1.get_pulls.assert_called_once_with()
        repo2.get_pulls.assert_called_once_with()
        mock_init.assert_has_calls([
            mock.call(repo1, 'pr1_1'),
            mock.call(repo1, 'pr1_2'),
            mock.call(repo1, 'pr1_3'),
            mock.call(repo2, 'pr2_1'),
            mock.call(repo2, 'pr2_2'),
        ])
        self.assertEqual(mock_init.call_count, 5)
        cb.assert_has_calls([
            mock.call(0, 2, repo1),
            mock.call(0, 2, repo1, result[:3]),
            mock.call(1, 2, repo2),
            mock.call(1, 2, repo2, result[3:]),
        ])
        self.assertEqual(cb.call_count, 4)

    @mock.patch.object(pulls.PullRequest, '_from_repos', return_value='pulls')
    def test_from_repo(self, mock_from_repos):
        gh = mock.Mock(**{'get_repo.return_value': 'repo'})

        result = pulls.PullRequest.from_repo(gh, 'spam')

        self.assertEqual(result, 'pulls')
        gh.get_repo.assert_called_once_with('spam')
        mock_from_repos.assert_called_once_with(['repo'], None)

    @mock.patch.object(pulls.PullRequest, '_from_repos', return_value='pulls')
    def test_from_repo_callback(self, mock_from_repos):
        gh = mock.Mock(**{'get_repo.return_value': 'repo'})

        result = pulls.PullRequest.from_repo(gh, 'spam', 'call')

        self.assertEqual(result, 'pulls')
        gh.get_repo.assert_called_once_with('spam')
        mock_from_repos.assert_called_once_with(['repo'], 'call')

    @mock.patch.object(pulls.PullRequest, '_from_repos', return_value='pulls')
    def test_from_organization(self, mock_from_repos):
        org = mock.Mock(**{'get_repos.return_value': ['repo1', 'repo2']})
        gh = mock.Mock(**{'get_organization.return_value': org})

        result = pulls.PullRequest.from_organization(gh, 'spam')

        self.assertEqual(result, 'pulls')
        gh.get_organization.assert_called_once_with('spam')
        org.get_repos.assert_called_once_with()
        mock_from_repos.assert_called_once_with(['repo1', 'repo2'], None)

    @mock.patch.object(pulls.PullRequest, '_from_repos', return_value='pulls')
    def test_from_organization_callback(self, mock_from_repos):
        org = mock.Mock(**{'get_repos.return_value': ['repo1', 'repo2']})
        gh = mock.Mock(**{'get_organization.return_value': org})

        result = pulls.PullRequest.from_organization(gh, 'spam', 'call')

        self.assertEqual(result, 'pulls')
        gh.get_organization.assert_called_once_with('spam')
        org.get_repos.assert_called_once_with()
        mock_from_repos.assert_called_once_with(['repo1', 'repo2'], 'call')

    @mock.patch.object(pulls.PullRequest, '_from_repos', return_value='pulls')
    def test_from_user(self, mock_from_repos):
        user = mock.Mock(**{'get_repos.return_value': ['repo1', 'repo2']})
        gh = mock.Mock(**{'get_user.return_value': user})

        result = pulls.PullRequest.from_user(gh, 'spam')

        self.assertEqual(result, 'pulls')
        gh.get_user.assert_called_once_with('spam')
        user.get_repos.assert_called_once_with()
        mock_from_repos.assert_called_once_with(['repo1', 'repo2'], None)

    @mock.patch.object(pulls.PullRequest, '_from_repos', return_value='pulls')
    def test_from_user_callback(self, mock_from_repos):
        user = mock.Mock(**{'get_repos.return_value': ['repo1', 'repo2']})
        gh = mock.Mock(**{'get_user.return_value': user})

        result = pulls.PullRequest.from_user(gh, 'spam', 'call')

        self.assertEqual(result, 'pulls')
        gh.get_user.assert_called_once_with('spam')
        user.get_repos.assert_called_once_with()
        mock_from_repos.assert_called_once_with(['repo1', 'repo2'], 'call')

    @mock.patch.object(pulls.PullRequest, '_from_repos', return_value='pulls')
    def test_from_all(self, mock_from_repos):
        gh = mock.Mock(**{'get_repos.return_value': ['repo1', 'repo2']})

        result = pulls.PullRequest.from_all(gh)

        self.assertEqual(result, 'pulls')
        gh.get_repos.assert_called_once_with()
        mock_from_repos.assert_called_once_with(['repo1', 'repo2'], None)

    @mock.patch.object(pulls.PullRequest, '_from_repos', return_value='pulls')
    def test_from_all_callback(self, mock_from_repos):
        gh = mock.Mock(**{'get_repos.return_value': ['repo1', 'repo2']})

        result = pulls.PullRequest.from_all(gh, 'call')

        self.assertEqual(result, 'pulls')
        gh.get_repos.assert_called_once_with()
        mock_from_repos.assert_called_once_with(['repo1', 'repo2'], 'call')

    def test_init(self):
        pr = pulls.PullRequest('repo', 'pr')

        self.assertEqual(pr._repo, 'repo')
        self.assertEqual(pr._pr, 'pr')
        self.assertEqual(pr._mergeable, None)

    def test_getattr(self):
        pr = pulls.PullRequest('repo', mock.Mock(attr='spam'))

        self.assertEqual(pr.attr, 'spam')

    def test_mergeable_cached(self):
        pr = pulls.PullRequest('repo', mock.Mock(mergeable='mergeable'))
        pr._mergeable = 'cached'

        self.assertEqual(pr.mergeable, 'cached')
        self.assertEqual(pr._mergeable, 'cached')

    def test_mergeable_uncached(self):
        pr = pulls.PullRequest('repo', mock.Mock(mergeable='mergeable'))

        self.assertEqual(pr.mergeable, 'mergeable')
        self.assertEqual(pr._mergeable, 'mergeable')

    def test_mergeable_clear(self):
        pr = pulls.PullRequest('repo', 'pr')
        pr._mergeable = 'cached'

        del pr.mergeable

        self.assertEqual(pr._mergeable, None)

    def test_repo(self):
        pr = pulls.PullRequest('repo', 'pr')

        self.assertEqual(pr.repo, 'repo')

    def test_pr(self):
        pr = pulls.PullRequest('repo', 'pr')

        self.assertEqual(pr.pr, 'pr')



# Copyright 2014 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the
#    License. You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing,
#    software distributed under the License is distributed on an "AS
#    IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#    express or implied. See the License for the specific language
#    governing permissions and limitations under the License.


class PullRequest(object):
    """
    Wrap a ``github.PullRequest.PullRequest`` object.  This provides
    some caching to reduce the number of round trips necessary when
    processing a list of pull requests.
    """

    @classmethod
    def _from_repos(cls, repos, repo_callback):
        """
        Given a list of repositories, builds and returns a list of all
        pull requests in those repositories.

        :param repos: A sequence of repositories.
        :param repo_callback: A callback to invoke for each repository
                              visited.  The callback will be called
                              twice.  The first time, it will be
                              passed the repository index, the total
                              length of the list of repositories, and
                              the ``github.Repository.Repository``
                              object; this call will be made prior to
                              retrieving the list of pull requests.
                              The second call will be made after
                              retrieving the list of pull requests,
                              and will include that list as the fourth
                              argument.

        :returns: A list of ``PullRequest`` objects.
        """

        repos = list(repos)
        pulls = []
        for idx, repo in enumerate(repos):
            # Emit a status update
            if repo_callback:
                repo_callback(idx, len(repos), repo)

            repo_pulls = [cls(repo, pr) for pr in repo.get_pulls()]

            # Emit a second status update with the pulls
            if repo_callback:
                repo_callback(idx, len(repos), repo, repo_pulls)

            pulls.extend(repo_pulls)

        return pulls

    @classmethod
    def from_repo(cls, gh, repo_name, repo_callback=None):
        """
        Retrieve all open pull requests from the named repository.

        :param gh: A ``github.Github`` handle.
        :param repo_name: The full name of the repository.
        :param repo_callback: A callback to invoke for each repository
                              visited.  The callback will be called
                              twice.  The first time, it will be
                              passed the repository index, the total
                              length of the list of repositories, and
                              the ``github.Repository.Repository``
                              object; this call will be made prior to
                              retrieving the list of pull requests.
                              The second call will be made after
                              retrieving the list of pull requests,
                              and will include that list as the fourth
                              argument.

        :returns: A list of ``PullRequest`` objects for each open pull
                  request against the named repository.  The list is
                  not sorted.
        """

        # This is pretty simple...
        return cls._from_repos([gh.get_repo(repo_name)], repo_callback)

    @classmethod
    def from_organization(cls, gh, org_name, repo_callback=None):
        """
        Retrieve all open pull requests from all repositories in a given
        organization.

        :param gh: A ``github.Github`` handle.
        :param org_name: The name of the organization.
        :param repo_callback: A callback to invoke for each repository
                              visited.  The callback will be called
                              twice.  The first time, it will be
                              passed the repository index, the total
                              length of the list of repositories, and
                              the ``github.Repository.Repository``
                              object; this call will be made prior to
                              retrieving the list of pull requests.
                              The second call will be made after
                              retrieving the list of pull requests,
                              and will include that list as the fourth
                              argument.

        :returns: A list of ``PullRequest`` objects for each open pull
                  request against all repositories in the named
                  organization.  The list is not sorted.
        """

        # First, get the organization
        org = gh.get_organization(org_name)

        # Now build and return the list of pull requests
        return cls._from_repos(org.get_repos(), repo_callback)

    @classmethod
    def from_user(cls, gh, user_name, repo_callback=None):
        """
        Retrieve all open pull requests from all repositories belonging to
        a given user.

        :param gh: A ``github.Github`` handle.
        :param user_name: The user login name.
        :param repo_callback: A callback to invoke for each repository
                              visited.  The callback will be called
                              twice.  The first time, it will be
                              passed the repository index, the total
                              length of the list of repositories, and
                              the ``github.Repository.Repository``
                              object; this call will be made prior to
                              retrieving the list of pull requests.
                              The second call will be made after
                              retrieving the list of pull requests,
                              and will include that list as the fourth
                              argument.

        :returns: A list of ``PullRequest`` objects for each open pull
                  request against all repositories belonging to the
                  named user.  The list is not sorted.
        """

        # First, get the user
        user = gh.get_user(user_name)

        # Now build and return the list of pull requests
        return cls._from_repos(user.get_repos(), repo_callback)

    @classmethod
    def from_all(cls, gh, repo_callback=None):
        """
        Retrieve all open pull requests from all repositories on Github.

        :param gh: A ``github.Github`` handle.
        :param repo_callback: A callback to invoke for each repository
                              visited.  The callback will be called
                              twice.  The first time, it will be
                              passed the repository index, the total
                              length of the list of repositories, and
                              the ``github.Repository.Repository``
                              object; this call will be made prior to
                              retrieving the list of pull requests.
                              The second call will be made after
                              retrieving the list of pull requests,
                              and will include that list as the fourth
                              argument.

        :returns: A list of ``PullRequest`` objects for each open pull
                  request against all repositories on Github which are
                  accessible to the user used when constructing
                  ``gh``.
        """

        # Build and return the list of all pull requests
        return cls._from_repos(gh.get_repos(), repo_callback)

    def __init__(self, repo, pr):
        """
        Initialize a ``PullRequest`` object.

        :param repo: The ``github.Repository.Repository`` object
                     describing the repository the pull request is
                     against.
        :param pr: The ``github.PullRequest.PullRequest`` object
                   describing the pull request.
        """

        self._repo = repo
        self._pr = pr

        self._mergeable = None

    def __getattr__(self, name):
        """
        Delegate attribute access to the underlying
        ``github.PullRequest.PullRequest`` object.

        :param name: The name of the desired attribute.

        :returns: The value of the desired attribute.
        """

        return getattr(self._pr, name)

    @property
    def mergeable(self):
        """
        Determine if the pull request is mergeable.  This is cached to
        inhibit round-tripping.  Use ``del x.mergeable`` to force a
        cache invalidation.
        """

        # Do we have the value cached?
        if self._mergeable is None:
            self._mergeable = self._pr.mergeable

        return self._mergeable

    @mergeable.deleter
    def mergeable(self):
        """
        Invalidate the caching of ``mergeable``.  This will force a round
        trip next time the ``mergeable`` property is accessed,
        yielding the most up-to-date value.
        """

        self._mergeable = None

    @property
    def repo(self):
        """
        Return the underlying ``github.Repository.Repository`` object that
        the pull request is against.
        """

        return self._repo

    @property
    def pr(self):
        """
        Return the underlying ``github.PullRequest.PullRequest`` object.
        """

        return self._pr



# Copyright 2014 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the
#    License. You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing,
#    software distributed under the License is distributed on an "AS
#    IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#    express or implied. See the License for the specific language
#    governing permissions and limitations under the License.



# Copyright 2014 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the
#    License. You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing,
#    software distributed under the License is distributed on an "AS
#    IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#    express or implied. See the License for the specific language
#    governing permissions and limitations under the License.

from __future__ import print_function

import argparse
import datetime
import getpass
import os
import sys

import cli_tools
import github

from tugboat import pulls


class RepoSummary(object):
    """
    A container for information about repositories.  This is used by
    ``report()`` to maintain a count of pull requests and mergeable
    pull requests for reporting in the final summary data.
    """

    def __init__(self, name):
        """
        Initialize a ``RepoSummary`` object.

        :param name: The full name of the repository.
        """

        self.name = name
        self.pulls = 0
        self.mergeable = 0

    def __iadd__(self, other):
        """
        Add the pull request to the summary.

        :param other: The pull request to add to the summary.

        :returns: The ``RepoSummary`` object.
        """

        # Count it
        self.pulls += 1

        # Is it mergeable?
        if other.mergeable:
            self.mergeable += 1

        return self


# This maps the target name used in an argument declaration to the
# routine used to find the open pull requests for that target
targets = {
    'repo': pulls.PullRequest.from_repo,
    'organization': pulls.PullRequest.from_organization,
    'user': pulls.PullRequest.from_user,
}


class RepoAction(argparse.Action):
    """
    An ``argparse.Action`` subclass used for command line arguments
    that are used in indicating which repositories to report open pull
    requests on.
    """

    def __init__(self, option_strings, dest, **kwargs):
        """
        Initialize a ``RepoAction`` object.

        :param option_strings: A list of option strings.
        :param dest: The target attribute to store the option values
                     in.
        :param target: The target the option specifies.  Must be one
                       of the keys from the ``targets`` dictionary.
                       Defaults to "repo" if not provided.
        """

        # Need the target information
        target = kwargs.pop('target', 'repo')

        # Initialize the Action
        super(RepoAction, self).__init__(option_strings, dest, **kwargs)

        # Save the target information
        self.target = target

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Called when encountering an argument bound to the ``RepoAction``
        object.

        :param parser: An ``argparse.ArgumentParser`` object.
        :param namespace: The ``argparse.Namespace`` into which to
                          store argument values.
        :param values: The values that were passed for the argument.
        :param option_string: The string used to invoke the option.
        """

        # Append the appropriate value to the namespace
        items = getattr(namespace, self.dest, [])
        items.append((self.target, values))
        setattr(namespace, self.dest, items)


@cli_tools.argument_group(
    'auth',
    title='Authentication-related Options',
    description='Options used to authenticate to Github.',
)
@cli_tools.argument(
    '--username', '-u',
    default=getpass.getuser(),
    help='Username for accessing the Github API.  Defaults to "%(default)s".',
    group='auth',
)
@cli_tools.argument(
    '--password', '-p',
    help='Password or personal access token for accessing the Github API.  '
    'If not provided, it will be prompted for.',
    group='auth',
)
@cli_tools.argument(
    '--github-url', '-g',
    default='https://api.github.com',
    help='API URL for accessing the Github API.  Defaults to "%(default)s".',
    group='auth',
)
@cli_tools.argument_group(
    'repo',
    title='Repositories to Report on',
    description='Options used to identify specific repositories to generate '
    'reports on.  These options may be used multiple times.',
)
@cli_tools.argument(
    '--repo', '-r',
    dest='repos',
    action=RepoAction,
    default=[],
    help='Identify a specific repository to generate a report for.  The '
    'full repository name must be provided, e.g., "<login>/<repo>".',
    group='repo',
)
@cli_tools.argument(
    '--user', '-U',
    dest='repos',
    action=RepoAction,
    help='Identify a user to generate a report for.  The user login name '
    'must be provided.  The report will include all visible repositories '
    'belonging to that user.',
    group='repo',
    target='user',
)
@cli_tools.argument(
    '--org', '-o',
    dest='repos',
    action=RepoAction,
    help='Identify an organization to generate a report for.  The '
    'organization name must be provided.  The report will include all '
    'visible repositories belonging to that user.',
    group='repo',
    target='organization',
)
@cli_tools.argument(
    '--output', '-O',
    default='-',
    help='Specify the file name the report should be emitted to.  If not '
    'provided, or if specified as "-", the report will be emitted to '
    'standard output.',
)
@cli_tools.argument(
    '--verbose', '-v',
    action='store_const',
    default=1,
    const=2,
    help='Request verbose output.  This will cause status messages to be '
    'emitted while producing the report.',
)
@cli_tools.argument(
    '--quiet', '-q',
    dest='verbose',
    action='store_const',
    const=0,
    help='Request quiet output.  This will suppress all status messages, '
    'emitting only the final report.',
)
@cli_tools.argument(
    '--debug', '-d',
    action='store_true',
    help='Enable debugging mode.  If errors occur, a more detailed output '
    'will be emitted.  This does not affect verbosity.'
)
def report(gh, repos, stream=sys.stdout, repo_callback=None):
    """
    Generate a report of all open pull requests on the specified
    repositories (see the "--repo", "--user", and "--org" options for
    how to specify repositories).

    :param gh: A ``github.Github`` handle for accessing the Github
               API.
    :param repos: A list of tuples specifying repositories to obtain
                  the report on.  For each element of the list, the
                  first element of the tuple is one of "repo", "user",
                  or "organization", and the second element is the
                  name of that repository, user, or organization,
                  respectively.
    :param stream: The output stream to receive the report.  Defaults
                   to ``sys.stdout``.
    :param repo_callback: A callback to invoke for each repository
                          visited.  The callback will be called twice.
                          The first time, it will be passed the
                          repository index, the total length of the
                          list of repositories, and the
                          ``github.Repository.Repository`` object;
                          this call will be made prior to retrieving
                          the list of pull requests.  The second call
                          will be made after retrieving the list of
                          pull requests, and will include that list as
                          the fourth argument.
    """

    start = datetime.datetime.now()

    # Build the list of pull requests
    pulls = []
    for target, name in repos:
        # Emit some status information
        if repo_callback:
            print(u'Looking up %s "%s"...' % (target, name),
                  file=sys.stderr)

        repo_pulls = targets[target](gh, name, repo_callback)

        pulls.extend(repo_pulls)

    # Now we need to sort the list of pulls...
    pulls.sort(key=lambda x: x.updated_at)

    # Emit one last piece of status information
    if repo_callback:
        print(u'Generating report...', file=sys.stderr)

    # Don't do anything if there are no pulls
    if not pulls:
        print(u"No open pull requests", file=stream)
        return

    # Emit a summary
    print(u"Open PRs: %d (%d mergeable); oldest last updated %s" %
          (len(pulls), sum(1 for pull in pulls if pull.mergeable),
           pulls[0].updated_at), file=stream)

    # Generate the report of pulls
    repos = {}
    for pull in pulls:
        print(u"\n"
              u"Pull request {pull.repo.full_name}#{pull.number}:\n"
              u"    URL: {pull.html_url}\n"
              u"    Merge {pull.head.label} -> {pull.base.label}\n"
              u"    Proposed {pull.created_at} by {username} "
              u"({pull.user.login})\n"
              u"    Last updated: {pull.updated_at}\n"
              u"    Mergeable: {mergeable}".format(
                  pull=pull,
                  mergeable=('yes' if pull.mergeable else 'no'),
                  username=(pull.user.name or '<unknown>')).encode('utf-8'),
              file=stream)

        # Add repository breakdown data
        repos.setdefault(pull.repo.full_name, RepoSummary(pull.repo.full_name))
        repos[pull.repo.full_name] += pull

    # Generate the repository breakdown
    print((u"\nRepositories with open pull requests: %d\n"
          "Breakdown by repository:" % len(repos)).encode('utf-8'),
          file=stream)
    for summary in sorted(repos.values(), key=lambda x: x.name):
        print((u"    Open PRs for %s: %d (%d mergeable)" %
               (summary.name, summary.pulls,
                summary.mergeable)).encode('utf-8'),
              file=stream)

    # Emit the time data
    end = datetime.datetime.now()
    print((u"\nReport generated in %s at %s" %
           (end - start, start)).encode('utf-8'),
          file=stream)


def _normal_callback(idx, count, repo, pulls=None):
    """
    A ``repo_callback`` callback to implement "normal" behavior.  This
    implementation emits a status line indicating the name of the
    repository being processed.

    :param idx: The index of the repository in the list being
                processed.
    :param count: The number of repositories in the list being
                  processed.
    :param repo: The repository being processed.
    :param pulls: The list of ``tugboat.pulls.PullRequest`` objects
                  for all open pull requests on the repository.
    """

    if pulls is None:
        print(u'Processing repository "%s" (%d/%d)...' %
              (repo.full_name, idx + 1, count), file=sys.stderr)


def _verbose_callback(idx, count, repo, pulls=None):
    """
    A ``repo_callback`` callback to implement "verbose" ("-v")
    behavior.  This implementation emits a status line indicating the
    name of the repository being processed, then indicates the total
    number of pull requests and counts the number of mergeable pull
    requests.

    :param idx: The index of the repository in the list being
                processed.
    :param count: The number of repositories in the list being
                  processed.
    :param repo: The repository being processed.
    :param pulls: The list of ``tugboat.pulls.PullRequest`` objects
                  for all open pull requests on the repository.
    """

    if pulls is None:
        print(u'Processing repository "%s" (%d/%d)...' %
              (repo.full_name, idx + 1, count), file=sys.stderr, end=' ')
    else:
        print(u'%d pulls' % len(pulls), file=sys.stderr, end=' ')
        print(u'(%d mergeable)' % sum(1 for pull in pulls if pull.mergeable),
              file=sys.stderr)


# Used to translate the verbosity level into an appropriate verbosity
# callback
verbosity = [None, _normal_callback, _verbose_callback]


@report.processor
def _process_report(args):
    """
    A ``cli_tools`` processor that adapts between the command line
    interface and the ``report()`` function.  The processor obtains a
    ``github.Github`` object, using the authentication data collected
    by the argument processor; it then selects the correct output
    stream and ``repo_callback`` function for the verbosity level.
    After ``report()`` returns, it ensures that the output stream is
    closed, if required.

    :param args: The ``argparse.Namespace`` object constructed by
                 ``cli_tools``.

    :returns: A ``cli_tools`` processor generator.
    """

    # Get the user's password
    password = args.password
    if not password:
        password = getpass.getpass(u'Password for %s> ' % args.username)

    # Create a github handle
    args.gh = github.Github(args.username, password, args.github_url)

    # Select the correct output stream
    if args.output == '-':
        args.stream = sys.stdout
        close = False
    else:
        args.stream = open(args.output, 'w')
        close = True

    # Select the correct verbosity
    args.repo_callback = verbosity[args.verbose]

    # Generate the report as requested
    try:
        yield
    finally:
        # Make sure the stream gets closed
        if close:
            args.stream.close()



# Copyright 2014 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the
#    License. You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing,
#    software distributed under the License is distributed on an "AS
#    IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#    express or implied. See the License for the specific language
#    governing permissions and limitations under the License.

import sys
import unittest

import mock
import six
from six.moves import builtins

from tugboat import reports


class RepoSummaryTest(unittest.TestCase):
    def test_init(self):
        result = reports.RepoSummary('repo')

        self.assertEqual(result.name, 'repo')
        self.assertEqual(result.pulls, 0)
        self.assertEqual(result.mergeable, 0)

    def test_iadd_unmergeable(self):
        summary = reports.RepoSummary('repo')
        pull = mock.Mock(mergeable=False)

        summary += pull

        self.assertEqual(summary.pulls, 1)
        self.assertEqual(summary.mergeable, 0)

    def test_iadd_mergeable(self):
        summary = reports.RepoSummary('repo')
        pull = mock.Mock(mergeable=True)

        summary += pull

        self.assertEqual(summary.pulls, 1)
        self.assertEqual(summary.mergeable, 1)


class RepoActionTest(unittest.TestCase):
    @mock.patch('argparse.Action.__init__', return_value=None)
    def test_init_no_target(self, mock_init):
        result = reports.RepoAction('strings', 'dest', a=1, b=2, c=3)

        self.assertEqual(result.target, 'repo')
        mock_init.assert_called_once_with('strings', 'dest', a=1, b=2, c=3)

    @mock.patch('argparse.Action.__init__', return_value=None)
    def test_init_with_target(self, mock_init):
        result = reports.RepoAction('strings', 'dest', a=1, b=2, c=3,
                                    target='organization')

        self.assertEqual(result.target, 'organization')
        mock_init.assert_called_once_with('strings', 'dest', a=1, b=2, c=3)

    @mock.patch('argparse.Action.__init__', return_value=None)
    def test_call(self, mock_init):
        action = reports.RepoAction('strings', 'dest', target='user')
        action.dest = 'dest'
        namespace = mock.Mock(spec=[])

        action('parser', namespace, 'spam')
        action('parser', namespace, 'foo')

        self.assertEqual(namespace.dest, [('user', 'spam'), ('user', 'foo')])


class ReportTest(unittest.TestCase):
    @mock.patch('datetime.datetime', mock.Mock(now=mock.Mock(side_effect=[
        1234567,
        1234569,
    ])))
    @mock.patch.dict(reports.targets, clear=True)
    @mock.patch.object(sys, 'stderr', six.StringIO())
    def test_basic(self):
        prs = {
            'repo1#1': mock.Mock(**{
                'user.name': None,
                'mergeable': False,
                'created_at': 10,
                'updated_at': 90,
            }),
            'repo1#2': mock.Mock(**{
                'user.name': None,
                'mergeable': True,
                'created_at': 20,
                'updated_at': 80,
            }),
            'repo2#1': mock.Mock(**{
                'user.name': 'spam',
                'mergeable': True,
                'created_at': 30,
                'updated_at': 70,
            }),
            'user1:repo3#1': mock.Mock(**{
                'user.name': None,
                'mergeable': True,
                'created_at': 40,
                'updated_at': 60,
            }),
            'user1:repo4#1': mock.Mock(**{
                'user.name': None,
                'mergeable': False,
                'created_at': 50,
                'updated_at': 50,
            }),
            'user2:repo5#1': mock.Mock(**{
                'user.name': None,
                'mergeable': True,
                'created_at': 60,
                'updated_at': 40,
            }),
            'org1:repo6#1': mock.Mock(**{
                'user.name': 'spam',
                'mergeable': True,
                'created_at': 70,
                'updated_at': 30,
            }),
            'org1:repo7#1': mock.Mock(**{
                'user.name': 'spam',
                'mergeable': True,
                'created_at': 80,
                'updated_at': 20,
            }),
            'org2:repo8#1': mock.Mock(**{
                'user.name': 'spam',
                'mergeable': True,
                'created_at': 90,
                'updated_at': 10,
            }),
        }
        for key, pr in prs.items():
            repo, number = key.split(':')[-1].split('#')
            pr.repo.full_name = repo
            pr.number = int(number)
            pr.html_url = 'https://github/%s/pull/%s' % (repo, number)
            pr.head.label = 'me:branch'
            pr.base.label = '%s:master' % repo
            pr.user.login = 'me'
        reports.targets = {
            'repo': mock.Mock(side_effect=lambda x, y, z: [
                pr for n, pr in prs.items() if n.startswith('%s#' % y)]),
            'user': mock.Mock(side_effect=lambda x, y, z: [
                pr for n, pr in prs.items() if n.startswith('%s:' % y)]),
            'organization': mock.Mock(side_effect=lambda x, y, z: [
                pr for n, pr in prs.items() if n.startswith('%s:' % y)]),
        }
        repos = [
            ('repo', 'repo1'),
            ('user', 'user1'),
            ('organization', 'org1'),
            ('repo', 'repo2'),
            ('user', 'user2'),
            ('organization', 'org2'),
        ]
        stream = six.StringIO()

        reports.report('gh', repos, stream, None)

        reports.targets['repo'].assert_has_calls([
            mock.call('gh', 'repo1', None),
            mock.call('gh', 'repo2', None),
        ])
        self.assertEqual(reports.targets['repo'].call_count, 2)
        reports.targets['user'].assert_has_calls([
            mock.call('gh', 'user1', None),
            mock.call('gh', 'user2', None),
        ])
        self.assertEqual(reports.targets['user'].call_count, 2)
        reports.targets['organization'].assert_has_calls([
            mock.call('gh', 'org1', None),
            mock.call('gh', 'org2', None),
        ])
        self.assertEqual(reports.targets['organization'].call_count, 2)
        self.assertEqual(
            stream.getvalue(),
            'Open PRs: 9 (7 mergeable); oldest last updated 10\n'
            '\n'
            'Pull request repo8#1:\n'
            '    URL: https://github/repo8/pull/1\n'
            '    Merge me:branch -> repo8:master\n'
            '    Proposed 90 by spam (me)\n'
            '    Last updated: 10\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo7#1:\n'
            '    URL: https://github/repo7/pull/1\n'
            '    Merge me:branch -> repo7:master\n'
            '    Proposed 80 by spam (me)\n'
            '    Last updated: 20\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo6#1:\n'
            '    URL: https://github/repo6/pull/1\n'
            '    Merge me:branch -> repo6:master\n'
            '    Proposed 70 by spam (me)\n'
            '    Last updated: 30\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo5#1:\n'
            '    URL: https://github/repo5/pull/1\n'
            '    Merge me:branch -> repo5:master\n'
            '    Proposed 60 by <unknown> (me)\n'
            '    Last updated: 40\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo4#1:\n'
            '    URL: https://github/repo4/pull/1\n'
            '    Merge me:branch -> repo4:master\n'
            '    Proposed 50 by <unknown> (me)\n'
            '    Last updated: 50\n'
            '    Mergeable: no\n'
            '\n'
            'Pull request repo3#1:\n'
            '    URL: https://github/repo3/pull/1\n'
            '    Merge me:branch -> repo3:master\n'
            '    Proposed 40 by <unknown> (me)\n'
            '    Last updated: 60\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo2#1:\n'
            '    URL: https://github/repo2/pull/1\n'
            '    Merge me:branch -> repo2:master\n'
            '    Proposed 30 by spam (me)\n'
            '    Last updated: 70\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo1#2:\n'
            '    URL: https://github/repo1/pull/2\n'
            '    Merge me:branch -> repo1:master\n'
            '    Proposed 20 by <unknown> (me)\n'
            '    Last updated: 80\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo1#1:\n'
            '    URL: https://github/repo1/pull/1\n'
            '    Merge me:branch -> repo1:master\n'
            '    Proposed 10 by <unknown> (me)\n'
            '    Last updated: 90\n'
            '    Mergeable: no\n'
            '\n'
            'Repositories with open pull requests: 8\n'
            'Breakdown by repository:\n'
            '    Open PRs for repo1: 2 (1 mergeable)\n'
            '    Open PRs for repo2: 1 (1 mergeable)\n'
            '    Open PRs for repo3: 1 (1 mergeable)\n'
            '    Open PRs for repo4: 1 (0 mergeable)\n'
            '    Open PRs for repo5: 1 (1 mergeable)\n'
            '    Open PRs for repo6: 1 (1 mergeable)\n'
            '    Open PRs for repo7: 1 (1 mergeable)\n'
            '    Open PRs for repo8: 1 (1 mergeable)\n'
            '\n'
            'Report generated in 2 at 1234567\n'
        )
        self.assertEqual(sys.stderr.getvalue(), '')

    @mock.patch('datetime.datetime', mock.Mock(now=mock.Mock(side_effect=[
        1234567,
        1234569,
    ])))
    @mock.patch.dict(reports.targets, clear=True)
    @mock.patch.object(sys, 'stderr', six.StringIO())
    def test_callback(self):
        prs = {
            'repo1#1': mock.Mock(**{
                'user.name': None,
                'mergeable': False,
                'created_at': 10,
                'updated_at': 90,
            }),
            'repo1#2': mock.Mock(**{
                'user.name': None,
                'mergeable': True,
                'created_at': 20,
                'updated_at': 80,
            }),
            'repo2#1': mock.Mock(**{
                'user.name': 'spam',
                'mergeable': True,
                'created_at': 30,
                'updated_at': 70,
            }),
            'user1:repo3#1': mock.Mock(**{
                'user.name': None,
                'mergeable': True,
                'created_at': 40,
                'updated_at': 60,
            }),
            'user1:repo4#1': mock.Mock(**{
                'user.name': None,
                'mergeable': False,
                'created_at': 50,
                'updated_at': 50,
            }),
            'user2:repo5#1': mock.Mock(**{
                'user.name': None,
                'mergeable': True,
                'created_at': 60,
                'updated_at': 40,
            }),
            'org1:repo6#1': mock.Mock(**{
                'user.name': 'spam',
                'mergeable': True,
                'created_at': 70,
                'updated_at': 30,
            }),
            'org1:repo7#1': mock.Mock(**{
                'user.name': 'spam',
                'mergeable': True,
                'created_at': 80,
                'updated_at': 20,
            }),
            'org2:repo8#1': mock.Mock(**{
                'user.name': 'spam',
                'mergeable': True,
                'created_at': 90,
                'updated_at': 10,
            }),
        }
        for key, pr in prs.items():
            repo, number = key.split(':')[-1].split('#')
            pr.repo.full_name = repo
            pr.number = int(number)
            pr.html_url = 'https://github/%s/pull/%s' % (repo, number)
            pr.head.label = 'me:branch'
            pr.base.label = '%s:master' % repo
            pr.user.login = 'me'
        reports.targets = {
            'repo': mock.Mock(side_effect=lambda x, y, z: [
                pr for n, pr in prs.items() if n.startswith('%s#' % y)]),
            'user': mock.Mock(side_effect=lambda x, y, z: [
                pr for n, pr in prs.items() if n.startswith('%s:' % y)]),
            'organization': mock.Mock(side_effect=lambda x, y, z: [
                pr for n, pr in prs.items() if n.startswith('%s:' % y)]),
        }
        repos = [
            ('repo', 'repo1'),
            ('user', 'user1'),
            ('organization', 'org1'),
            ('repo', 'repo2'),
            ('user', 'user2'),
            ('organization', 'org2'),
        ]
        stream = six.StringIO()

        reports.report('gh', repos, stream, 'callback')

        reports.targets['repo'].assert_has_calls([
            mock.call('gh', 'repo1', 'callback'),
            mock.call('gh', 'repo2', 'callback'),
        ])
        self.assertEqual(reports.targets['repo'].call_count, 2)
        reports.targets['user'].assert_has_calls([
            mock.call('gh', 'user1', 'callback'),
            mock.call('gh', 'user2', 'callback'),
        ])
        self.assertEqual(reports.targets['user'].call_count, 2)
        reports.targets['organization'].assert_has_calls([
            mock.call('gh', 'org1', 'callback'),
            mock.call('gh', 'org2', 'callback'),
        ])
        self.assertEqual(reports.targets['organization'].call_count, 2)
        self.assertEqual(
            stream.getvalue(),
            'Open PRs: 9 (7 mergeable); oldest last updated 10\n'
            '\n'
            'Pull request repo8#1:\n'
            '    URL: https://github/repo8/pull/1\n'
            '    Merge me:branch -> repo8:master\n'
            '    Proposed 90 by spam (me)\n'
            '    Last updated: 10\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo7#1:\n'
            '    URL: https://github/repo7/pull/1\n'
            '    Merge me:branch -> repo7:master\n'
            '    Proposed 80 by spam (me)\n'
            '    Last updated: 20\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo6#1:\n'
            '    URL: https://github/repo6/pull/1\n'
            '    Merge me:branch -> repo6:master\n'
            '    Proposed 70 by spam (me)\n'
            '    Last updated: 30\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo5#1:\n'
            '    URL: https://github/repo5/pull/1\n'
            '    Merge me:branch -> repo5:master\n'
            '    Proposed 60 by <unknown> (me)\n'
            '    Last updated: 40\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo4#1:\n'
            '    URL: https://github/repo4/pull/1\n'
            '    Merge me:branch -> repo4:master\n'
            '    Proposed 50 by <unknown> (me)\n'
            '    Last updated: 50\n'
            '    Mergeable: no\n'
            '\n'
            'Pull request repo3#1:\n'
            '    URL: https://github/repo3/pull/1\n'
            '    Merge me:branch -> repo3:master\n'
            '    Proposed 40 by <unknown> (me)\n'
            '    Last updated: 60\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo2#1:\n'
            '    URL: https://github/repo2/pull/1\n'
            '    Merge me:branch -> repo2:master\n'
            '    Proposed 30 by spam (me)\n'
            '    Last updated: 70\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo1#2:\n'
            '    URL: https://github/repo1/pull/2\n'
            '    Merge me:branch -> repo1:master\n'
            '    Proposed 20 by <unknown> (me)\n'
            '    Last updated: 80\n'
            '    Mergeable: yes\n'
            '\n'
            'Pull request repo1#1:\n'
            '    URL: https://github/repo1/pull/1\n'
            '    Merge me:branch -> repo1:master\n'
            '    Proposed 10 by <unknown> (me)\n'
            '    Last updated: 90\n'
            '    Mergeable: no\n'
            '\n'
            'Repositories with open pull requests: 8\n'
            'Breakdown by repository:\n'
            '    Open PRs for repo1: 2 (1 mergeable)\n'
            '    Open PRs for repo2: 1 (1 mergeable)\n'
            '    Open PRs for repo3: 1 (1 mergeable)\n'
            '    Open PRs for repo4: 1 (0 mergeable)\n'
            '    Open PRs for repo5: 1 (1 mergeable)\n'
            '    Open PRs for repo6: 1 (1 mergeable)\n'
            '    Open PRs for repo7: 1 (1 mergeable)\n'
            '    Open PRs for repo8: 1 (1 mergeable)\n'
            '\n'
            'Report generated in 2 at 1234567\n'
        )
        self.assertEqual(sys.stderr.getvalue(),
                         'Looking up repo "repo1"...\n'
                         'Looking up user "user1"...\n'
                         'Looking up organization "org1"...\n'
                         'Looking up repo "repo2"...\n'
                         'Looking up user "user2"...\n'
                         'Looking up organization "org2"...\n'
                         'Generating report...\n')

    @mock.patch('datetime.datetime', mock.Mock(now=mock.Mock(side_effect=[
        1234567,
        1234569,
    ])))
    @mock.patch.dict(reports.targets, clear=True)
    @mock.patch.object(sys, 'stderr', six.StringIO())
    def test_empty(self):
        stream = six.StringIO()

        reports.report('gh', [], stream, 'callback')

        self.assertEqual(stream.getvalue(), 'No open pull requests\n')
        self.assertEqual(sys.stderr.getvalue(), 'Generating report...\n')


class NormalCallbackTest(unittest.TestCase):
    @mock.patch.object(sys, 'stderr', six.StringIO())
    def test_no_pulls(self):
        repo = mock.Mock(full_name='repo')

        reports._normal_callback(1, 3, repo)

        self.assertEqual(sys.stderr.getvalue(),
                         'Processing repository "repo" (2/3)...\n')

    @mock.patch.object(sys, 'stderr', six.StringIO())
    def test_with_pulls(self):
        repo = mock.Mock(full_name='repo')
        pulls = [
            mock.Mock(mergeable=True),
            mock.Mock(mergeable=False),
            mock.Mock(mergeable=True),
        ]

        reports._normal_callback(1, 3, repo, pulls)

        self.assertEqual(sys.stderr.getvalue(), '')


class VerboseCallbackTest(unittest.TestCase):
    @mock.patch.object(sys, 'stderr', six.StringIO())
    def test_no_pulls(self):
        repo = mock.Mock(full_name='repo')

        reports._verbose_callback(1, 3, repo)

        self.assertEqual(sys.stderr.getvalue(),
                         'Processing repository "repo" (2/3)... ')

    @mock.patch.object(sys, 'stderr', six.StringIO())
    def test_with_pulls(self):
        repo = mock.Mock(full_name='repo')
        pulls = [
            mock.Mock(mergeable=True),
            mock.Mock(mergeable=False),
            mock.Mock(mergeable=True),
        ]

        reports._verbose_callback(1, 3, repo, pulls)

        self.assertEqual(sys.stderr.getvalue(), '3 pulls (2 mergeable)\n')


class ProcessReportTest(unittest.TestCase):
    @mock.patch('getpass.getpass', return_value='prompted')
    @mock.patch('github.Github', return_value='gh')
    @mock.patch.object(builtins, 'open')
    @mock.patch('sys.stdout', mock.Mock())
    def test_basic(self, mock_open, mock_Github, mock_getpass):
        args = mock.Mock(username='username', password='password',
                         github_url='github_url', output='-',
                         verbose=0)

        gen = reports._process_report(args)
        next(gen)

        self.assertEqual(args.gh, 'gh')
        self.assertEqual(args.stream, sys.stdout)
        self.assertEqual(args.repo_callback, None)
        self.assertFalse(mock_getpass.called)
        mock_Github.assert_called_once_with(
            'username', 'password', 'github_url')
        self.assertFalse(mock_open.called)
        self.assertFalse(sys.stdout.close.called)

        try:
            next(gen)
        except StopIteration:
            pass
        else:
            self.fail('Failed to end iteration')

        self.assertFalse(sys.stdout.close.called)

    @mock.patch('getpass.getpass', return_value='prompted')
    @mock.patch('github.Github', return_value='gh')
    @mock.patch.object(builtins, 'open')
    @mock.patch('sys.stdout', mock.Mock())
    def test_prompt(self, mock_open, mock_Github, mock_getpass):
        args = mock.Mock(username='username', password=None,
                         github_url='github_url', output='-',
                         verbose=0)

        gen = reports._process_report(args)
        next(gen)

        self.assertEqual(args.gh, 'gh')
        self.assertEqual(args.stream, sys.stdout)
        self.assertEqual(args.repo_callback, None)
        mock_getpass.assert_called_once_with('Password for username> ')
        mock_Github.assert_called_once_with(
            'username', 'prompted', 'github_url')
        self.assertFalse(mock_open.called)
        self.assertFalse(sys.stdout.close.called)

        try:
            next(gen)
        except StopIteration:
            pass
        else:
            self.fail('Failed to end iteration')

        self.assertFalse(sys.stdout.close.called)

    @mock.patch('getpass.getpass', return_value='prompted')
    @mock.patch('github.Github', return_value='gh')
    @mock.patch.object(builtins, 'open')
    @mock.patch('sys.stdout', mock.Mock())
    def test_output(self, mock_open, mock_Github, mock_getpass):
        args = mock.Mock(username='username', password='password',
                         github_url='github_url', output='output',
                         verbose=0)

        gen = reports._process_report(args)
        next(gen)

        self.assertEqual(args.gh, 'gh')
        self.assertEqual(args.stream, mock_open.return_value)
        self.assertEqual(args.repo_callback, None)
        self.assertFalse(mock_getpass.called)
        mock_Github.assert_called_once_with(
            'username', 'password', 'github_url')
        mock_open.assert_called_once_with('output', 'w')
        self.assertFalse(mock_open.return_value.close.called)

        try:
            next(gen)
        except StopIteration:
            pass
        else:
            self.fail('Failed to end iteration')

        mock_open.return_value.close.assert_called_once_with()

    @mock.patch('getpass.getpass', return_value='prompted')
    @mock.patch('github.Github', return_value='gh')
    @mock.patch.object(builtins, 'open')
    @mock.patch('sys.stdout', mock.Mock())
    def test_verbosity_normal(self, mock_open, mock_Github, mock_getpass):
        args = mock.Mock(username='username', password='password',
                         github_url='github_url', output='-',
                         verbose=1)

        gen = reports._process_report(args)
        next(gen)

        self.assertEqual(args.gh, 'gh')
        self.assertEqual(args.stream, sys.stdout)
        self.assertEqual(args.repo_callback, reports._normal_callback)
        self.assertFalse(mock_getpass.called)
        mock_Github.assert_called_once_with(
            'username', 'password', 'github_url')
        self.assertFalse(mock_open.called)
        self.assertFalse(sys.stdout.close.called)

        try:
            next(gen)
        except StopIteration:
            pass
        else:
            self.fail('Failed to end iteration')

        self.assertFalse(sys.stdout.close.called)

    @mock.patch('getpass.getpass', return_value='prompted')
    @mock.patch('github.Github', return_value='gh')
    @mock.patch.object(builtins, 'open')
    @mock.patch('sys.stdout', mock.Mock())
    def test_verbosity_verbose(self, mock_open, mock_Github, mock_getpass):
        args = mock.Mock(username='username', password='password',
                         github_url='github_url', output='-',
                         verbose=2)

        gen = reports._process_report(args)
        next(gen)

        self.assertEqual(args.gh, 'gh')
        self.assertEqual(args.stream, sys.stdout)
        self.assertEqual(args.repo_callback, reports._verbose_callback)
        self.assertFalse(mock_getpass.called)
        mock_Github.assert_called_once_with(
            'username', 'password', 'github_url')
        self.assertFalse(mock_open.called)
        self.assertFalse(sys.stdout.close.called)

        try:
            next(gen)
        except StopIteration:
            pass
        else:
            self.fail('Failed to end iteration')

        self.assertFalse(sys.stdout.close.called)
